import os
import shutil
import tempfile

import nitrate
import pytest
from fmf import Tree
from requre import RequreTestCase
from ruamel.yaml import YAML

import tmt.base
import tmt.cli
import tmt.log
from tests import CliRunner
from tmt.utils import ConvertError, Path

# Prepare path to examples
TEST_DIR = Path(__file__).parent


class Base(RequreTestCase):
    EXAMPLES = TEST_DIR / "data" / "nitrate"

    def setUp(self):
        super().setUp()
        self.tmpdir = Path(tempfile.mkdtemp(prefix=str(TEST_DIR)))
        shutil.copytree(self.EXAMPLES, self.tmpdir, dirs_exist_ok=True)
        self.cwd = os.getcwd()
        self.runner_output = None

        # Disable nitrate cache
        nitrate.set_cache_level(0)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        os.chdir(self.cwd)
        super().tearDown()
        if hasattr(
                self.runner_output,
                "exit_code") and self.runner_output.exit_code != 0:
            print("Return code:", self.runner_output.exit_code)
            print("Output:", self.runner_output.output)
            print("Exception:", self.runner_output.exception)


# General test plan for this component is: TP#29309
class NitrateExport(Base):

    def test_create(self):
        fmf_node = Tree(self.tmpdir).find("/new_testcase")
        assert "extra-nitrate" not in fmf_node.data

        os.chdir(self.tmpdir / "new_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main,
            ["test", "export", "--how", "nitrate", "--ignore-git-validation",
             "--create", "--general", "--append-summary", "."],
            catch_exceptions=False)
        # Reload the node data to see if it appears there
        fmf_node = Tree(self.tmpdir).find("/new_testcase")
        assert "extra-nitrate" in fmf_node.data

    def test_create_dryrun(self):
        fmf_node_before = Tree(self.tmpdir).find("/new_testcase")
        assert "extra-nitrate" not in fmf_node_before.data

        os.chdir(self.tmpdir / "new_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main,
            ["test", "export", "--how", "nitrate", "--ignore-git-validation",
             "--create", "--dry", "--general", "--append-summary", "."],
            catch_exceptions=False)
        fmf_node = Tree(self.tmpdir).find("/new_testcase")
        assert "extra-nitrate" not in fmf_node.data
        assert fmf_node_before.data == fmf_node.data
        assert "summary: tmt /new_testcase - This i" in self.runner_output.output

    def test_existing(self):
        fmf_node = Tree(self.tmpdir).find("/existing_testcase")
        assert fmf_node.data["extra-nitrate"] == "TC#0609686"

        os.chdir(self.tmpdir / "existing_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main,
            ["test", "export", "--how", "nitrate", "--ignore-git-validation",
             "--create", "--general", "--append-summary", "."],
            catch_exceptions=False)
        fmf_node = Tree(self.tmpdir).find("/existing_testcase")

    def test_existing_dryrun(self):
        fmf_node = Tree(self.tmpdir).find("/existing_dryrun_testcase")
        assert fmf_node.data["extra-nitrate"] == "TC#0609686"

        os.chdir(self.tmpdir / "existing_dryrun_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main,
            ["test", "export", "--how", "nitrate", "--ignore-git-validation",
             "--debug", "--dry", "--general", "--bugzilla", "--append-summary", "."],
            catch_exceptions=False)
        assert "summary: tmt /existing_dryrun_testcase - ABCDEF" in self.runner_output.output

    def test_existing_release_dryrun(self):
        fmf_node = Tree(self.tmpdir).find("/existing_dryrun_release_testcase")
        assert fmf_node.data["extra-nitrate"] == "TC#0609686"

        os.chdir(self.tmpdir / "existing_dryrun_release_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(tmt.cli.main,
                                           ["test",
                                            "export",
                                            "--how",
                                            "nitrate",
                                            "--debug",
                                            "--dry",
                                            "--ignore-git-validation",
                                            "--general",
                                            "--bugzilla",
                                            "--link-runs",
                                            "--append-summary",
                                            "."],
                                           catch_exceptions=False)
        assert "summary: tmt /existing_dryrun_release_testcase - ABCDEF" in \
               self.runner_output.output
        assert "Linked to general plan 'TP#28164 - tmt / General'" in self.runner_output.output
        assert "Link to plan 'TP#31698" in self.runner_output.output
        assert "Link to run 'TR#425023" in self.runner_output.output

    def test_coverage_bugzilla(self):
        fmf_node = Tree(self.tmpdir).find("/existing_testcase")
        assert fmf_node.data["extra-nitrate"] == "TC#0609686"

        os.chdir(self.tmpdir / "existing_testcase")
        runner = CliRunner()
        self.runner_output = runner.invoke(tmt.cli.main,
                                           ["test",
                                            "export",
                                            "--how",
                                            "nitrate",
                                            "--ignore-git-validation",
                                            "--bugzilla",
                                            "--append-summary",
                                            "."],
                                           catch_exceptions=False)
        assert self.runner_output.exit_code == 0

    def test_missing_user_dryrun(self):
        os.chdir(self.tmpdir / "existing_testcase_missing_user")
        runner = CliRunner()
        with pytest.raises(ConvertError):
            self.runner_output = runner.invoke(
                tmt.cli.main,
                ["test", "export", "--how", "nitrate",
                 "--debug", "--dry", "."],
                catch_exceptions=False)

    def test_export_blocked_by_validation(self):
        os.chdir(self.tmpdir / "validation")
        fmf_node = Tree(self.tmpdir).find("/validation")
        with fmf_node as data:
            data['test'] = 'echo hello world'
        runner = CliRunner()
        with pytest.raises(ConvertError) as error:
            self.runner_output = runner.invoke(
                tmt.cli.main,
                ["test", "export", "--nitrate", "--debug", "--dry", "--append-summary", "."],
                catch_exceptions=False)
        assert "Uncommitted changes" in str(error.value)

    def test_export_forced_validation(self):
        os.chdir(self.tmpdir / "validation")
        fmf_node = Tree(self.tmpdir).find("/validation")
        with fmf_node as data:
            data['extra-nitrate'] = 'TC#599605'

        runner = CliRunner()

        self.runner_output = runner.invoke(
            tmt.cli.main,
            ["test", "export", "--nitrate", "--debug", "--ignore-git-validation",
             "--append-summary", "."],
            catch_exceptions=False)

        assert "Exporting regardless 'Uncommitted changes" in self.runner_output.output


class NitrateImport(Base):

    def test_import_manual_confirmed(self):
        runner = CliRunner()
        # TODO: import does not respect --root param anyhow (could)
        self.runner_output = runner.invoke(
            tmt.cli.main,
            ['-vvvvdddd', '--root', self.tmpdir / "import_case",
             "test", "import", "--no-general", "--nitrate", "--manual", "--case=609704"],
            catch_exceptions=False)
        assert self.runner_output.exit_code == 0
        assert "Importing the 'Imported_Test_Case'" in self.runner_output.output
        assert "test case found 'TC#0609704'" in self.runner_output.output
        assert "Metadata successfully stored into" in self.runner_output.output
        filename = next(
            filter(
                lambda x: "Metadata successfully stored into" in x
                          and "main.fmf" in x,
                self.runner_output.output.splitlines())).split("'")[1]
        # /home/jscotka/git/tmt/Manual/Imported_Test_Case/main.fmf
        # TODO: not possible to specify, where to store data,
        # it always creates Manual subdir, I do not want it.
        assert "/Manual/Imported_Test_Case/main.fmf" in filename
        assert Path(filename).exists()
        with open(Path(filename)) as file:
            yaml = YAML(typ='safe')
            out = yaml.load(file)
            assert "Tier1" in out["tag"]
            assert "tmt_test_component" in out["component"]

    def test_import_manual_proposed(self):
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main, ['--root', self.tmpdir / "import_case", "test",
                           "import", "--no-general", "--nitrate", "--manual", "--case=609705"],
            catch_exceptions=False)
        assert self.runner_output.exit_code == 0
        # TODO: This is strange, expect at least some output in
        # case there is proper case, just case is not CONFIRMED
        # I can imagine also e.g. at least raise error but not pass,
        # with no output
        assert self.runner_output.output.strip() == ""
        fmf_node = Tree(self.tmpdir).find("/import_case")
        assert fmf_node is None


class NitrateImportAutomated(Base):
    test_md_content = """# Setup
Do this and that to setup the environment.

# Test

## Step
Step one.

Step two.

Step three.

## Expect
Expect one.

Expect two

Expect three.

# Cleanup
This is a breakdown.
"""
    main_fmf_content = """summary: Simple smoke test
description: |
    Just run 'tmt --help' to make sure the binary is sane.
    This is really that simple. Nothing more here. Really.
contact: Petr Šplíchal <psplicha@redhat.com>
component:
- tmt
test: ./runtest.sh
framework: beakerlib
require:
- fmf
recommend:
- tmt
duration: 5m
enabled: true
tag: []
link:
-   relates: https://bugzilla.redhat.com/show_bug.cgi?id=12345
-   relates: https://bugzilla.redhat.com/show_bug.cgi?id=1234567
adjust:
-   because: comment
    enabled: false
    when: distro == rhel-4, rhel-5
    continue: false
-   environment:
        PHASES: novalgrind
    when: arch == s390x
    continue: false
extra-nitrate: TC#0609926
extra-summary: /tmt/integration
extra-task: /tmt/integration
"""

    def test_basic(self):
        os.chdir(self.tmpdir / "import_case_automated")
        files = os.listdir()
        assert "Makefile" in files
        assert "main.fmf" not in files
        assert "test.md" not in files
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main, [
                "test", "import", "--nitrate"], catch_exceptions=False)
        assert self.runner_output.exit_code == 0
        files = os.listdir()
        assert "Makefile" in files
        assert "test.md" in files
        with open("test.md") as file:
            assert self.test_md_content in file.read()
        assert "main.fmf" in files
        with open("main.fmf") as file:
            yaml = YAML(typ='safe')
            generated = yaml.load(file)
            referenced = yaml.load(self.main_fmf_content)
            assert generated == referenced

    def test_old_relevancy(self):
        os.chdir(self.tmpdir / "import_old_relevancy")
        files = os.listdir()
        assert files == ["Makefile"]
        runner = CliRunner()
        self.runner_output = runner.invoke(
            tmt.cli.main, [
                "test", "import", "--nitrate", "--no-general"], catch_exceptions=False)
        assert self.runner_output.exit_code == 0

        tree_f36_intel = tmt.Tree(
            logger=tmt.log.Logger.create(),
            path='.',
            fmf_context={
                'distro': ['fedora-36'],
                'arch': ['x86_64']})

        found_tests = tree_f36_intel.tests(names=['/import_old_relevancy'])
        assert len(found_tests) == 1
        test = found_tests[0]
        assert test.enabled
        assert test.environment == {"ARCH": "not arch"}
        assert test.node.get("extra-nitrate") == "TC#0545993"

        tree_f35_intel = tmt.Tree(
            logger=tmt.log.Logger.create(),
            path='.',
            fmf_context={
                'distro': ['fedora-35'],
                'arch': ['x86_64']})

        found_tests = tree_f35_intel.tests(names=['/import_old_relevancy'])
        assert len(found_tests) == 1
        test = found_tests[0]
        assert not test.enabled
        # second rule is ignored if the order is correctly transferred
        assert test.environment == {}
        assert test.node.get("extra-nitrate") == "TC#0545993"



"""
ReST rendering.

Package provides primitives for ReST rendering used mainly for CLI
help texts.
"""

import functools
from typing import Optional

import click
import docutils.frontend
import docutils.nodes
import docutils.parsers.rst
import docutils.utils

import tmt.log
from tmt.log import Logger
from tmt.utils import GeneralError

#: Special string representing a new-line in the stack of rendered
#: paragraphs.
NL = ''


class RestVisitor(docutils.nodes.NodeVisitor):
    """
    Custom renderer of docutils nodes.

    See :py:class:`docutils.nodes.NodeVisitor` for details, but the
    functionality is fairly simple: for each node type, a pair of
    methods is expected, ``visit_$NODE_TYPE`` and ``depart_$NODE_TYPE``.
    As the visitor class iterates over nodes in the document,
    corresponding methods are called. These methods render the given
    node, filling "rendered paragraphs" list with rendered strings.
    """

    def __init__(self, document: docutils.nodes.document, logger: Logger) -> None:
        super().__init__(document)

        self.logger = logger
        self.debug = functools.partial(logger.debug, level=4, topic=tmt.log.Topic.HELP_RENDERING)
        self.log_visit = functools.partial(
            logger.debug, 'visit', level=4, topic=tmt.log.Topic.HELP_RENDERING)
        self.log_departure = functools.partial(
            logger.debug, 'depart', level=4, topic=tmt.log.Topic.HELP_RENDERING)

        #: Collects all rendered paragraps - text, blocks, lists, etc.
        self._rendered_paragraphs: list[str] = []
        #: Collect components of a single paragraph - sentences, literals,
        #: list items, etc.
        self._rendered_paragraph: list[str] = []

        self.in_literal_block: bool = False
        self.in_note: bool = False
        self.in_warning: bool = False

        #: Used by rendering of nested blocks, e.g. paragraphs positioned
        #: as list items.
        self._indent: int = 0
        self._text_prefix: Optional[str] = None

    @property
    def rendered(self) -> str:
        """ Return the rendered document as a single string """

        # Drop any trailing empty lines
        while self._rendered_paragraphs and self._rendered_paragraphs[-1] == NL:
            self._rendered_paragraphs.pop(-1)

        return '\n'.join(self._rendered_paragraphs)

    def _emit(self, s: str) -> None:
        """ Add a string to the paragraph being rendered """

        self._rendered_paragraph.append(s)

    def _emit_paragraphs(self, paragraphs: list[str]) -> None:
        """ Add new rendered paragraphs """

        self._rendered_paragraphs += paragraphs

    def flush(self) -> None:
        """ Finalize rendering of the current paragraph """

        if not self._rendered_paragraph:
            self.nl()

        else:
            self._emit_paragraphs([''.join(self._rendered_paragraph)])
            self._rendered_paragraph = []

    def nl(self) -> None:
        """ Render a new, empty line """

        # To simplify the implementation, this is merging of multiple
        # empty lines into one. Rendering of nodes than does not have
        # to worry about an empty line already being on the stack.
        if self._rendered_paragraphs and self._rendered_paragraphs[-1] != NL:
            self._emit_paragraphs([NL])

    # Simple logging for nodes that have no effect
    def _noop_visit(self, node: docutils.nodes.Node) -> None:
        self.log_visit(str(node))

    def _noop_departure(self, node: docutils.nodes.Node) -> None:
        self.log_departure(str(node))

    # Node renderers
    visit_document = _noop_visit

    def depart_document(self, node: docutils.nodes.document) -> None:
        self.log_departure(str(node))

        self.flush()

    def visit_paragraph(self, node: docutils.nodes.paragraph) -> None:
        self.log_visit(str(node))

        if isinstance(node.parent, docutils.nodes.list_item):
            if self._text_prefix:
                self._emit(self._text_prefix)
                self._text_prefix = None

            else:
                self._emit(' ' * self._indent)

        elif self.in_note:
            self._emit(click.style('NOTE: ', fg='blue', bold=True))
            return

        elif self.in_warning:
            self._emit(click.style('WARNING: ', fg='yellow', bold=True))
            return

    def depart_paragraph(self, node: docutils.nodes.paragraph) -> None:
        self.log_departure(str(node))

        # Top-level paragraphs should be followed by an empty line to
        # prevent paragraphs sticking together. Only the top-level ones
        # though, we do not want empty lines after every paragraph-like
        # string, because a lot of nodes are also paragraphs.
        if isinstance(node.parent, docutils.nodes.document):
            self.nl()

        self.flush()

    def visit_Text(self, node: docutils.nodes.Text) -> None:  # noqa: N802
        self.log_visit(str(node))

        if isinstance(node.parent, docutils.nodes.literal):
            return

        if self.in_literal_block:
            return

        if self.in_note:
            self._emit(click.style(node.astext(), fg='blue'))

            return

        if self.in_warning:
            self._emit(click.style(node.astext(), fg='yellow'))

            return

        self._emit(node.astext())

    depart_Text = _noop_departure  # noqa: N815

    def visit_literal(self, node: docutils.nodes.literal) -> None:
        self.log_visit(str(node))

        self._emit(click.style(node.astext(), fg='green'))

    depart_literal = _noop_departure

    def visit_literal_block(self, node: docutils.nodes.literal_block) -> None:
        self.log_visit(str(node))

        self.flush()

        fg: str = 'cyan'

        if 'yaml' in node.attributes['classes']:
            pass

        elif 'shell' in node.attributes['classes']:
            fg = 'yellow'

        self._emit_paragraphs([
            f'    {click.style(line, fg=fg)}' for line in node.astext().splitlines()
            ])

        self.in_literal_block = True

    def depart_literal_block(self, node: docutils.nodes.literal_block) -> None:
        self.log_departure(str(node))

        self.in_literal_block = False

        self.nl()

    def visit_bullet_list(self, node: docutils.nodes.bullet_list) -> None:
        self.log_visit(str(node))

        self.nl()

    def depart_bullet_list(self, node: docutils.nodes.bullet_list) -> None:
        self.log_departure(str(node))

        self.nl()

    def visit_list_item(self, node: docutils.nodes.list_item) -> None:
        self.log_visit(str(node))

        self._text_prefix = '* '
        self._indent += 2

    def depart_list_item(self, node: docutils.nodes.list_item) -> None:
        self.log_departure(str(node))

        self._indent -= 2

    visit_inline = _noop_visit
    depart_inline = _noop_departure

    visit_reference = _noop_visit
    depart_reference = _noop_departure

    def visit_note(self, node: docutils.nodes.note) -> None:
        self.log_visit(str(node))

        self.nl()
        self.in_note = True

    def depart_note(self, node: docutils.nodes.note) -> None:
        self.log_departure(str(node))

        self.in_note = False
        self.nl()

    def visit_warning(self, node: docutils.nodes.warning) -> None:
        self.log_visit(str(node))

        self.nl()
        self.in_warning = True

    def depart_warning(self, node: docutils.nodes.warning) -> None:
        self.log_departure(str(node))

        self.in_warning = False
        self.nl()

    def unknown_visit(self, node: docutils.nodes.Node) -> None:
        raise GeneralError(f"Unhandled ReST node '{node}'.")

    def unknown_departure(self, node: docutils.nodes.Node) -> None:
        raise GeneralError(f"Unhandled ReST node '{node}'.")


def parse_rst(text: str) -> docutils.nodes.document:
    """ Parse a ReST document into docutils tree of nodes """

    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(components=components).get_default_values()
    document = docutils.utils.new_document('<rst-doc>', settings=settings)

    parser.parse(text, document)

    return document


def render_rst(text: str, logger: Logger) -> str:
    """ Render a ReST document """

    document = parse_rst(text)
    visitor = RestVisitor(document, logger)

    document.walkabout(visitor)

    return visitor.rendered



import dataclasses
from collections.abc import Sequence
from typing import Optional

import tmt
import tmt.log
import tmt.steps
import tmt.steps.report
from tmt.result import BaseResult, CheckResult, Result, SubCheckResult, SubResult
from tmt.steps.execute import TEST_OUTPUT_FILENAME
from tmt.utils import Path, field

# How much test and test check info should be shifted to the right in the output.
# We want tests to be shifted by one extra level, with their checks shifted by
# yet another level.
TEST_SHIFT = 1
CHECK_SHIFT = 2
SUBRESULT_SHIFT = CHECK_SHIFT
SUBRESULT_CHECK_SHIFT = SUBRESULT_SHIFT + 1


@dataclasses.dataclass
class ReportDisplayData(tmt.steps.report.ReportStepData):
    display_guest: str = field(
        default='auto',
        option='--display-guest',
        metavar='auto|always|never',
        choices=['auto', 'always', 'never'],
        help="""
             When to display full guest name in report: when more than a single guest was involved
             (default), always, or never.
             """)


@tmt.steps.provides_method('display')
class ReportDisplay(tmt.steps.report.ReportPlugin[ReportDisplayData]):
    """
    Show test results on the terminal.

    Give a concise summary of test results directly on the terminal.
    List individual test results in verbose mode.
    """

    _data_class = ReportDisplayData

    def details(self, result: tmt.Result, verbosity: int, display_guest: bool) -> None:
        """
        Print result details.

        :param result: a test result to display.
        :param verbosity: how verbose should the report be. Generally equal to
            number of  ``--verbose``/``-v`` options given on command line.
            For ``1``, display only test name and its result, ``2`` will add
            log paths, and ``3`` or more would show the test output as well.
        :param display_guest: if set, guest multihost name would be part of the
            report.
        """

        def _display_log_info(log: Path, shift: int) -> None:
            """ Display info about a single result log """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            self.verbose(
                log.name,
                self.step.plan.execute.workdir / log,
                color='yellow',
                shift=shift)

        def _display_log_content(log: Path, shift: int) -> None:
            """ Display content of a single result log """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            self.verbose(
                'content',
                self.read(self.step.plan.execute.workdir / log),
                color='yellow',
                shift=shift)

        def display_outcome(result: BaseResult) -> None:
            """ Display a single result outcome """

            if isinstance(result, SubCheckResult):
                self.verbose(
                    f'{result.show()} ({result.event.value} check)',
                    shift=SUBRESULT_CHECK_SHIFT)

            elif isinstance(result, CheckResult):
                self.verbose(
                    f'{result.show()} ({result.event.value} check)',
                    shift=CHECK_SHIFT)

            elif isinstance(result, SubResult):
                self.verbose(result.show(), shift=SUBRESULT_SHIFT)

            elif isinstance(result, Result):
                self.verbose(result.show(display_guest=display_guest), shift=TEST_SHIFT)

        def display_log_info(result: BaseResult) -> None:
            """ Display info about result logs """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            shift = (TEST_SHIFT + 1) if isinstance(result, Result) else (CHECK_SHIFT + 1)

            for log in result.log:
                _display_log_info(log, shift)

        def display_log_content(result: BaseResult) -> None:
            """ Display content of interesting result logs """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            shift = (TEST_SHIFT + 1) if isinstance(result, Result) else (CHECK_SHIFT + 1)

            for log in result.log:
                _display_log_info(log, shift)

                if log.name == TEST_OUTPUT_FILENAME:
                    _display_log_content(log, shift)

        def display_subresults(results: Sequence[BaseResult]) -> None:
            """ Display subresults, checks and subresult checks """

            for subresult in results:
                display_outcome(subresult)

                if verbosity > 2:
                    display_log_content(subresult)

                elif verbosity > 1:
                    display_log_info(subresult)

                # Recursively show also all the results of subresult checks
                if isinstance(subresult, SubResult):
                    display_subresults(subresult.check)

        # Always show the result outcome
        display_outcome(result)

        # With verbosity increased to `-vvv` or more, display content of the main test log
        if verbosity > 2:
            display_log_content(result)

        # With verbosity increased to `-vv`, display the list of logs
        elif verbosity > 1:
            display_log_info(result)

        display_subresults(result.check)
        display_subresults(result.subresult)

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """ Discover available tests """
        super().go(logger=logger)
        # Show individual test results only in verbose mode
        if not self.verbosity_level:
            return

        if self.data.display_guest == 'always':
            display_guest = True

        elif self.data.display_guest == 'never':
            display_guest = False

        else:
            seen_guests = {
                result.guest.name
                for result in self.step.plan.execute.results()
                }

            display_guest = len(seen_guests) > 1

        for result in self.step.plan.execute.results():
            self.details(result, self.verbosity_level, display_guest)



import dataclasses
from typing import Any, Optional, cast

import fmf
import fmf.utils

import tmt
import tmt.log
import tmt.steps
import tmt.steps.prepare
import tmt.utils
from tmt.result import PhaseResult
from tmt.steps import safe_filename
from tmt.steps.provision import Guest
from tmt.utils import ShellScript, field

PREPARE_WRAPPER_FILENAME = 'tmt-prepare-wrapper.sh'


@dataclasses.dataclass
class PrepareShellData(tmt.steps.prepare.PrepareStepData):
    script: list[ShellScript] = field(
        default_factory=list,
        option=('-s', '--script'),
        multiple=True,
        metavar='SCRIPT',
        help='Shell script to be executed. Can be used multiple times.',
        normalize=tmt.utils.normalize_shell_script_list,
        serialize=lambda scripts: [str(script) for script in scripts],
        unserialize=lambda serialized: [ShellScript(script) for script in serialized]
        )

    # ignore[override] & cast: two base classes define to_spec(), with conflicting
    # formal types.
    def to_spec(self) -> dict[str, Any]:  # type: ignore[override]
        data = cast(dict[str, Any], super().to_spec())
        data['script'] = [str(script) for script in self.script]

        return data


@tmt.steps.provides_method('shell')
class PrepareShell(tmt.steps.prepare.PreparePlugin[PrepareShellData]):
    """
    Prepare guest using shell (Bash) scripts.

    Run various commands and scripts on the guest:

    .. code-block:: yaml

        prepare:
            how: shell
            script:
              - sudo dnf install -y 'dnf-command(copr)'
              - sudo dnf copr enable -y psss/tmt
              - sudo dnf install -y tmt
    """

    _data_class = PrepareShellData

    def go(
            self,
            *,
            guest: 'Guest',
            environment: Optional[tmt.utils.Environment] = None,
            logger: tmt.log.Logger) -> list[PhaseResult]:
        """ Prepare the guests """
        results = super().go(guest=guest, environment=environment, logger=logger)

        environment = environment or tmt.utils.Environment()

        # Give a short summary
        overview = fmf.utils.listed(self.data.script, 'script')
        logger.info('overview', f'{overview} found', 'green')

        workdir = self.step.plan.worktree
        assert workdir is not None  # narrow type

        if not self.is_dry_run:
            topology = tmt.steps.Topology(self.step.plan.provision.guests())
            topology.guest = tmt.steps.GuestTopology(guest)

            environment.update(
                topology.push(
                    dirpath=workdir,
                    guest=guest,
                    logger=logger,
                    filename_base=safe_filename(tmt.steps.TEST_TOPOLOGY_FILENAME_BASE, self, guest)
                    ))

        prepare_wrapper_filename = safe_filename(PREPARE_WRAPPER_FILENAME, self, guest)
        prepare_wrapper_path = workdir / prepare_wrapper_filename

        logger.debug('prepare wrapper', prepare_wrapper_path, level=3)

        # Execute each script on the guest (with default shell options)
        for script in self.data.script:
            logger.verbose('script', script, 'green')
            script_with_options = tmt.utils.ShellScript(f'{tmt.utils.SHELL_OPTIONS}; {script}')
            self.write(prepare_wrapper_path, str(script_with_options), 'w')
            if not self.is_dry_run:
                prepare_wrapper_path.chmod(0o755)
            guest.push(
                source=prepare_wrapper_path,
                destination=prepare_wrapper_path,
                options=["-s", "-p", "--chmod=755"])
            command: ShellScript
            if guest.become and not guest.facts.is_superuser:
                command = tmt.utils.ShellScript(f'sudo -E {prepare_wrapper_path}')
            else:
                command = tmt.utils.ShellScript(f'{prepare_wrapper_path}')
            guest.execute(
                command=command,
                cwd=workdir,
                env=environment)

        return results



import dataclasses
import os
import re
from time import time
from typing import TYPE_CHECKING, Any, Optional, overload

import requests

import tmt.log
import tmt.steps.report
from tmt.result import ResultOutcome
from tmt.utils import field, yaml_to_dict

if TYPE_CHECKING:
    from tmt._compat.typing import TypeAlias

JSON: 'TypeAlias' = Any


def _flag_env_to_default(option: str, default: bool) -> bool:
    env_var = 'TMT_PLUGIN_REPORT_REPORTPORTAL_' + option.upper()
    if env_var not in os.environ:
        return default
    return bool(os.getenv(env_var) == '1')


@overload
def _str_env_to_default(option: str, default: None) -> Optional[str]:
    pass


@overload
def _str_env_to_default(option: str, default: str) -> str:
    pass


def _str_env_to_default(option: str, default: Optional[str]) -> Optional[str]:
    env_var = 'TMT_PLUGIN_REPORT_REPORTPORTAL_' + option.upper()
    if env_var not in os.environ or os.getenv(env_var) is None:
        return default
    return str(os.getenv(env_var))


def _filter_invalid_chars(data: str) -> str:
    return re.sub(
        '[^\u0020-\uD7FF\u0009\u000A\u000D\uE000-\uFFFD\U00010000-\U0010FFFF]+',
        '',
        data)


@dataclasses.dataclass
class ReportReportPortalData(tmt.steps.report.ReportStepData):

    url: Optional[str] = field(
        option="--url",
        metavar="URL",
        default=_str_env_to_default('url', None),
        help="The URL of the ReportPortal instance where the data should be sent to.")

    token: Optional[str] = field(
        option="--token",
        metavar="TOKEN",
        default=_str_env_to_default('token', None),
        help="The token to use for upload to the ReportPortal instance (from the user profile).")

    project: Optional[str] = field(
        option="--project",
        metavar="PROJECT_NAME",
        default=_str_env_to_default('project', None),
        help="Name of the project into which the results should be uploaded.")

    launch: Optional[str] = field(
        option="--launch",
        metavar="LAUNCH_NAME",
        default=_str_env_to_default('launch', None),
        help="""
           Set the launch name, otherwise name of the plan is used by default.
           Should be defined with 'suite-per-plan' option or it will be named after the first plan.
           """)

    launch_description: Optional[str] = field(
        option="--launch-description",
        metavar="DESCRIPTION",
        default=_str_env_to_default('launch_description', None),
        help="""
             Pass the description for ReportPortal launch with 'suite-per-plan' option
             or append the original (plan summary) with additional info.
             Appends test description with 'upload-to-launch/suite' options.
             """)

    launch_per_plan: bool = field(
        option="--launch-per-plan",
        default=_flag_env_to_default('launch_per_plan', False),
        is_flag=True,
        help="Mapping launch per plan, creating one or more launches with no suite structure.")

    suite_per_plan: bool = field(
        option="--suite-per-plan",
        default=_flag_env_to_default('suite_per_plan', False),
        is_flag=True,
        help="""
             Mapping suite per plan, creating one launch and continuous uploading suites into it.
             Recommended to use with 'launch' and 'launch-description' options.
             Can be used with 'upload-to-launch' option for an additional upload of new suites.
             """)

    upload_to_launch: Optional[str] = field(
        option="--upload-to-launch",
        metavar="LAUNCH_ID",
        default=_str_env_to_default('upload_to_launch', None),
        help="""
           Pass the launch ID for an additional test/suite upload to an existing launch. ID can be
           found in the launch URL. Keep the launch structure with options 'launch/suite-per-plan'.
           To upload specific info into description see also 'launch-description'.
           """)

    upload_to_suite: Optional[str] = field(
        option="--upload-to-suite",
        metavar="SUITE_ID",
        default=_str_env_to_default('upload_to_suite', None),
        help="""
             Pass the suite ID for an additional test upload to a suite
             within an existing launch. ID can be found in the suite URL.
             To upload specific info into description see also 'launch-description'.
             """)

    launch_rerun: bool = field(
        option="--launch-rerun",
        default=_flag_env_to_default('launch_rerun', False),
        is_flag=True,
        help="""
             Rerun the last launch based on its name and unique test paths to create Retry item
             with a new version per each test. Supported in 'suite-per-plan' structure only.
             """)

    defect_type: Optional[str] = field(
        option="--defect-type",
        metavar="DEFECT_NAME",
        default=_str_env_to_default('defect_type', None),
        help="""
             Pass the defect type to be used for failed test, which is defined in the project
             (e.g. 'Idle'). 'To Investigate' is used by default.
             """)

    exclude_variables: str = field(
        option="--exclude-variables",
        metavar="PATTERN",
        default=_str_env_to_default('exclude_variables', "^TMT_.*"),
        help="""
             Regular expression for excluding environment variables
             from reporting to ReportPortal ('^TMT_.*' used by default).
             Parameters in ReportPortal get filtered out by the pattern
             to prevent overloading and to preserve the history aggregation
             for ReportPortal item if tmt id is not provided.
             """)

    api_version: str = field(
        option="--api-version",
        metavar="VERSION",
        default=_str_env_to_default('api_version', "v1"),
        help="Override the default reportportal API version (v1).")

    artifacts_url: Optional[str] = field(
        metavar="ARTIFACTS_URL",
        option="--artifacts-url",
        default=_str_env_to_default('artifacts_url',
                                    os.getenv('TMT_REPORT_ARTIFACTS_URL')),
        help="Link to test artifacts provided for report plugins.")

    launch_url: Optional[str] = None
    launch_uuid: Optional[str] = None
    suite_uuid: Optional[str] = None
    test_uuids: dict[int, str] = field(
        default_factory=dict
        )


@tmt.steps.provides_method("reportportal")
class ReportReportPortal(tmt.steps.report.ReportPlugin[ReportReportPortalData]):
    """
    Report test results to a ReportPortal instance via API.

    For communication with Report Portal API is necessary to provide
    following options:

    * token for authentication
    * url of the ReportPortal instance
    * project name

    In addition to command line options it's possible to use environment
    variables:

    .. code-block:: bash

        export TMT_PLUGIN_REPORT_REPORTPORTAL_${MY_OPTION}=${MY_VALUE}

    Assuming the URL and token are provided by the environment variables,
    the plan config can look like this:

    .. code-block:: yaml

        report:
            how: reportportal
            project: baseosqe

        context:
            ...

        environment:
            ...

    Where the context and environment sections must be filled with
    corresponding data in order to report context as attributes
    (arch, component, distro, trigger, compose, etc.) and
    environment variables as parameters in the Item Details.

    Other reported fmf data are summary, id, web link and contact per
    test.

    Two types of data structures are supported for reporting to ReportPortal:

    * 'launch-per-plan' mapping (default) that results in launch-test structure.
    * 'suite-per-plan' mapping that results in launch-suite-test structure.

    Supported report use cases:

    * Report a new run in launch-suite-test or launch-test structure
    * Report an additional rerun with 'launch-rerun' option and same launch name (->Retry items)
      or by reusing the run and reporting with 'again' option (->append logs)
    * To see plan progress, discover and report an empty (IDLE) run
      and reuse the run for execution and updating the report with 'again' option
    * Report contents of a new run to an existing launch via the URL ID in three ways:
      tests to launch, suites to launch and tests to suite.
    """

    _data_class = ReportReportPortalData

    TMT_TO_RP_RESULT_STATUS = {
        ResultOutcome.PASS: "PASSED",
        ResultOutcome.FAIL: "FAILED",
        ResultOutcome.ERROR: "FAILED",
        ResultOutcome.WARN: "FAILED",
        ResultOutcome.INFO: "SKIPPED"
        }

    def handle_response(self, response: requests.Response) -> None:
        """ Check the endpoint response and raise an exception if needed """

        self.debug("Response code from the endpoint", response.status_code)
        self.debug("Message from the endpoint", response.text)

        if not response.ok:
            raise tmt.utils.ReportError(
                f"Received non-ok status code {response.status_code} "
                f"from ReportPortal: {response.text}")

    def check_options(self) -> None:
        """ Check options for known troublesome combinations """

        if self.data.launch_per_plan and self.data.suite_per_plan:
            self.warn(
                "The options '--launch-per-plan' and '--suite-per-plan' are mutually exclusive. "
                "Default option '--launch-per-plan' is used.")
            self.data.suite_per_plan = False

        if self.data.launch_rerun and (self.data.upload_to_launch or self.data.upload_to_suite):
            self.warn("Unexpected option combination: "
                      "'--launch-rerun' is ignored when uploading additional tests.")

        if not self.data.suite_per_plan and self.data.launch_rerun:
            self.warn("Unexpected option combination: '--launch-rerun' "
                      "may cause an unexpected behaviour with launch-per-plan structure")

    def time(self) -> str:
        return str(int(time() * 1000))

    def get_headers(self) -> dict[str, str]:
        return {"Authorization": "bearer " + str(self.data.token),
                "accept": "*/*",
                "Content-Type": "application/json"}

    def get_url(self) -> str:
        return f"{self.data.url}/api/{self.data.api_version}/{self.data.project}"

    def construct_launch_attributes(self, suite_per_plan: bool,
                                    attributes: list[dict[str, str]]) -> list[dict[str, str]]:
        if not suite_per_plan or not self.step.plan.my_run:
            return attributes.copy()

        # Get common attributes across the plans
        merged_plans = [{key: value[0] for key, value in plan._fmf_context.items()}
                        for plan in self.step.plan.my_run.plans]
        result_dict = merged_plans[0]
        for current_plan in merged_plans[1:]:
            tmp_dict = {}
            for key, value in current_plan.items():
                if key in result_dict and result_dict[key] == value:
                    tmp_dict[key] = value
            result_dict = tmp_dict
        return [{'key': key, 'value': value} for key, value in result_dict.items()]

    def get_defect_type_locator(self, session: requests.Session,
                                defect_type: Optional[str]) -> str:
        if not defect_type:
            return "ti001"

        response = self.rp_api_get(session, "settings")
        defect_types = yaml_to_dict(response.text).get("subTypes")
        if not defect_types:
            return "ti001"

        groups_to_search = ['TO_INVESTIGATE', 'NO_DEFECT',
                            'SYSTEM_ISSUE', 'AUTOMATION_BUG', 'PRODUCT_BUG']
        for group_name in groups_to_search:
            defect_types_list = defect_types[group_name]
            dt_tmp = [dt['locator'] for dt in defect_types_list
                      if dt['longName'].lower() == defect_type.lower()]
            dt_locator = dt_tmp[0] if dt_tmp else None
            if dt_locator:
                break
        if not dt_locator:
            raise tmt.utils.ReportError(f"Defect type '{defect_type}' "
                                        "is not be defined in the project {self.data.project}")
        self.verbose("defect_type", defect_type, color="cyan", shift=1)
        return str(dt_locator)

    def rp_api_get(self, session: requests.Session, path: str) -> requests.Response:
        response = session.get(url=f"{self.get_url()}/{path}",
                               headers=self.get_headers())
        self.handle_response(response)
        return response

    def rp_api_post(self, session: requests.Session, path: str, json: JSON) -> requests.Response:
        response = session.post(url=f"{self.get_url()}/{path}",
                                headers=self.get_headers(),
                                json=json)
        self.handle_response(response)
        return response

    def rp_api_put(self, session: requests.Session, path: str, json: JSON) -> requests.Response:
        response = session.put(url=f"{self.get_url()}/{path}",
                               headers=self.get_headers(),
                               json=json)
        self.handle_response(response)
        return response

    def append_description(self, curr_description: str) -> str:
        """ Extend text with the launch description (if provided) """
        if self.data.launch_description:
            if curr_description:
                curr_description += "<br>" + self.data.launch_description
            else:
                curr_description = self.data.launch_description
        return curr_description

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Report test results to the endpoint

        Create a ReportPortal launch and its test items,
        fill it with all parts needed and report the logs.
        """

        super().go(logger=logger)

        if not self.data.url:
            raise tmt.utils.ReportError("No ReportPortal endpoint url provided.")
        self.data.url = self.data.url.rstrip("/")

        if not self.data.project:
            raise tmt.utils.ReportError("No ReportPortal project provided.")

        if not self.data.token:
            raise tmt.utils.ReportError("No ReportPortal token provided.")

        if not self.step.plan.my_run:
            raise tmt.utils.ReportError("No run data available.")

        self.check_options()

        launch_time = self.time()

        # Support for idle tests
        executed = bool(self.step.plan.execute.results())
        if executed:
            # launch time should be the earliest start time of all plans
            launch_time = min([r.start_time or self.time()
                               for r in self.step.plan.execute.results()])

        # Create launch, suites (if "--suite_per_plan") and tests;
        # or report to existing launch/suite if its id is given
        suite_per_plan = self.data.suite_per_plan
        launch_per_plan = self.data.launch_per_plan
        if not launch_per_plan and not suite_per_plan:
            launch_per_plan = True      # by default

        suite_id = self.data.upload_to_suite
        launch_id = self.data.upload_to_launch

        suite_uuid = self.data.suite_uuid
        launch_uuid = self.data.launch_uuid
        additional_upload = suite_id or launch_id or launch_uuid
        is_the_first_plan = self.step.plan == self.step.plan.my_run.plans[0]
        if not launch_uuid and suite_per_plan and not is_the_first_plan:
            rp_phases = list(self.step.plan.my_run.plans[0].report.phases(ReportReportPortal))
            if rp_phases:
                launch_uuid = rp_phases[0].data.launch_uuid

        create_test = not self.data.test_uuids
        create_suite = suite_per_plan and not (suite_uuid or suite_id)
        create_launch = not (launch_uuid or launch_id or suite_uuid or suite_id)

        launch_name = self.data.launch or self.step.plan.name
        suite_name = ""
        launch_url = ""

        launch_rerun = self.data.launch_rerun
        envar_pattern = self.data.exclude_variables or "$^"
        defect_type = self.data.defect_type or ""

        attributes = [
            {'key': key, 'value': value[0]}
            for key, value in self.step.plan._fmf_context.items()]
        launch_attributes = self.construct_launch_attributes(suite_per_plan, attributes)

        if suite_per_plan:
            launch_description = self.data.launch_description or ""
            suite_description = self.append_description(self.step.plan.summary or "")
        else:
            launch_description = self.step.plan.summary or ""
            launch_description = self.append_description(launch_description)
            suite_description = ""

        # Check whether artifacts URL has been provided
        if self.data.artifacts_url:
            launch_description += f"<br>{self.data.artifacts_url}"
            suite_description += f"<br>{self.data.artifacts_url}"

        # Communication with RP instance
        with tmt.utils.retry_session(status_forcelist=(
                429,   # Too Many Requests
                500,   # Internal Server Error
                502,   # Bad Gateway
                503,   # Service Unavailable
                504,   # Gateway Timeout
                )) as session:

            if create_launch:

                # Create a launch
                self.info("launch", launch_name, color="cyan")
                response = self.rp_api_post(
                    session=session,
                    path="launch",
                    json={"name": launch_name,
                          "description": launch_description,
                          "attributes": launch_attributes,
                          "startTime": launch_time,
                          "rerun": launch_rerun})
                launch_uuid = yaml_to_dict(response.text).get("id")

            else:
                # Get the launch_uuid or info to log
                if suite_id:
                    response = self.rp_api_get(session, f"item/{suite_id}")
                    suite_uuid = yaml_to_dict(response.text).get("uuid")
                    suite_name = str(yaml_to_dict(response.text).get("name"))
                    launch_id = yaml_to_dict(response.text).get("launchId")

                if launch_id:
                    response = self.rp_api_get(session, f"launch/{launch_id}")
                    launch_uuid = yaml_to_dict(response.text).get("uuid")

            if launch_uuid and not launch_id:
                response = self.rp_api_get(session, f"launch/uuid/{launch_uuid}")
                launch_id = yaml_to_dict(response.text).get("id")

            # Print the launch info
            if not create_launch:
                launch_name = yaml_to_dict(response.text).get("name") or ""
                self.verbose("launch", launch_name, color="green")
                self.verbose("id", launch_id, "yellow", shift=1)

            assert launch_uuid is not None
            self.verbose("uuid", launch_uuid, "yellow", shift=1)
            self.data.launch_uuid = launch_uuid

            launch_url = f"{self.data.url}/ui/#{self.data.project}/launches/all/{launch_id}"

            if create_suite:
                # Create a suite
                suite_name = self.step.plan.name
                self.info("suite", suite_name, color="cyan")
                response = self.rp_api_post(
                    session=session,
                    path="item",
                    json={"name": suite_name,
                          "description": suite_description,
                          "attributes": attributes,
                          "startTime": launch_time,
                          "launchUuid": launch_uuid,
                          "type": "suite"})
                suite_uuid = yaml_to_dict(response.text).get("id")
                assert suite_uuid is not None

            elif suite_name:
                self.info("suite", suite_name, color="green")
                self.verbose("id", suite_id, "yellow", shift=1)

            if suite_uuid:
                self.verbose("uuid", suite_uuid, "yellow", shift=1)
                self.data.suite_uuid = suite_uuid

            for result, test in self.step.plan.execute.results_for_tests(
                    self.step.plan.discover.tests()):
                test_time = self.time()
                test_name = None
                test_description = ''
                test_link = None
                test_id = None
                env_vars = None

                item_attributes = attributes.copy()
                if result:
                    serial_number = result.serial_number
                    test_name = result.name
                    test_time = result.start_time or self.time()
                    # for guests, save their primary address
                    if result.guest.primary_address:
                        item_attributes.append({
                            'key': 'guest_primary_address',
                            'value': result.guest.primary_address})
                    # for multi-host tests store also provision name and role
                    if result.guest.name != 'default-0':
                        item_attributes.append(
                            {'key': 'guest_name', 'value': result.guest.name})
                    if result.guest.role:
                        item_attributes.append(
                            {'key': 'guest_role', 'value': result.guest.role})

                # update RP item with additional attributes if test details are available
                if test:
                    serial_number = test.serial_number
                    if not test_name:
                        test_name = test.name
                    if test.contact:
                        item_attributes.append({"key": "contact", "value": test.contact[0]})
                    if test.summary:
                        test_description = test.summary
                    if test.web_link():
                        test_link = test.web_link()
                    if test.id:
                        test_id = test.id
                    env_vars = [
                        {'key': key, 'value': value}
                        for key, value in test.environment.items()
                        if not re.search(envar_pattern, key)]

                if create_test:
                    if ((self.data.upload_to_launch and launch_per_plan)
                            or self.data.upload_to_suite):
                        test_description = self.append_description(test_description)

                    # Create a test item
                    self.info("test", test_name, color="cyan")
                    response = self.rp_api_post(
                        session=session,
                        path=f"item{f'/{suite_uuid}' if suite_uuid else ''}",
                        json={"name": test_name,
                              "description": test_description,
                              "attributes": item_attributes,
                              "parameters": env_vars,
                              "codeRef": test_link,
                              "launchUuid": launch_uuid,
                              "type": "step",
                              "testCaseId": test_id,
                              "startTime": test_time})
                    item_uuid = yaml_to_dict(response.text).get("id")
                    assert item_uuid is not None
                    self.verbose("uuid", item_uuid, "yellow", shift=1)
                    self.data.test_uuids[serial_number] = item_uuid
                else:
                    item_uuid = self.data.test_uuids[serial_number]
                # Support for idle tests
                status = "SKIPPED"
                if result:
                    # For each log
                    for index, log_path in enumerate(result.log):
                        try:
                            log = self.step.plan.execute.read(log_path)
                        except tmt.utils.FileError:
                            continue

                        level = "INFO" if log_path == result.log[0] else "TRACE"
                        status = self.TMT_TO_RP_RESULT_STATUS[result.result]

                        # Upload log
                        response = self.rp_api_post(
                            session=session,
                            path="log/entry",
                            json={"message": _filter_invalid_chars(log),
                                  "itemUuid": item_uuid,
                                  "launchUuid": launch_uuid,
                                  "level": level,
                                  "time": result.end_time})

                        # Write out failures
                        if index == 0 and status == "FAILED":
                            message = _filter_invalid_chars(result.failures(log))
                            response = self.rp_api_post(
                                session=session,
                                path="log/entry",
                                json={"message": message,
                                      "itemUuid": item_uuid,
                                      "launchUuid": launch_uuid,
                                      "level": "ERROR",
                                      "time": result.end_time})

                    test_time = result.end_time or self.time()

                # Finish the test item
                response = self.rp_api_put(
                    session=session,
                    path=f"item/{item_uuid}",
                    json={
                        "launchUuid": launch_uuid,
                        "endTime": test_time,
                        "status": status,
                        "issue": {
                            "issueType": self.get_defect_type_locator(session, defect_type)}})
                launch_time = test_time

            if create_suite:
                # Finish the test suite
                response = self.rp_api_put(
                    session=session,
                    path=f"item{f'/{suite_uuid}' if suite_uuid else ''}",
                    json={
                        "launchUuid": launch_uuid,
                        "endTime": launch_time})

            is_the_last_plan = self.step.plan == self.step.plan.my_run.plans[-1]
            if is_the_last_plan:
                self.data.defect_type = None

            if ((launch_per_plan or (suite_per_plan and is_the_last_plan))
                    and not additional_upload):
                # Finish the launch
                response = self.rp_api_put(
                    session=session,
                    path=f"launch/{launch_uuid}/finish",
                    json={"endTime": launch_time})
                launch_url = str(yaml_to_dict(response.text).get("link"))

            assert launch_url is not None
            self.info("url", launch_url, "magenta")
            self.data.launch_url = launch_url



import logging
import operator
import re
from collections.abc import Iterable
from typing import Any, Callable

import _pytest.logging
import pytest

from tmt.utils import remove_color


class PatternMatching:
    def __init__(self, pattern: str, method: str) -> None:
        self.pattern = pattern
        self._compiled_pattern = re.compile(pattern)
        self.method = getattr(self._compiled_pattern, method)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: "{self.pattern}">'


class MATCH(PatternMatching):
    """
    Wrap a string with this class, to use it as a regular expression when matching log records.

    :py:class:`SEARCH` applies to any place within the string, while ``MATCH`` must match from the
    beginning of the string.

    .. code-block:: python

       assert_log(message=MATCH('an exception .+ was raised'))
    """

    def __init__(self, pattern: str) -> None:
        super().__init__(pattern, 'match')


class SEARCH(PatternMatching):
    """
    Wrap a string with this class, to use it as a regular expression when searching for log
    records.

    ``SEARCH`` applies to any place within the string, while :py:class:`MATCH` must match from the
    beginning of the string.

    .. code-block:: python

       assert_log(message=SEARCH('an exception .+ was raised'))
    """

    def __init__(self, pattern: str) -> None:
        super().__init__(pattern, 'search')


def _assert_log(
        caplog: _pytest.logging.LogCaptureFixture,
        evaluator: Callable[[Iterable[Any]], bool] = any,
        remove_colors: bool = False,
        not_present: bool = False,
        **tests: Any
        ) -> None:
    """
    Assert log contains a record - logged message - with given properties. Those are specified as
    keyword parameters: :py:class:`logging.LogRecords` properties are allowed names, parameter
    values are the expected values.

    .. code-block:: python

       assert_log(message='everything went well', levelno=logging.INFO)
       assert_log(message='things broke down', levelno=logging.ERROR)
       assert_log(message=MATCH('user .+ logged in'), levelno=logging.INFO)

    :param caplog: Pytest's `caplog` fixture.
    :param evaluator: a callable reducing a given list of booleans into a single boolean. It is
        used to evaluate whether the search for matching record was successful: each record is
        tested, and results of these per-record tests are passed to `evaluator` for the final
        decision.
    """

    # We are given field_name=expected_value pairs, but we also want to be open to other binary
    # operators, like "field_name matches pattern". To protect the actual matching from aspects of
    # different possible operators, we will convert the "tests" into basic building blocks: a
    # field name, a callable accepting two parameters, and the given (expected) value. With these,
    # we can reduce the matching into functions calls without worrying what functions we work with.

    operators: list[tuple[Callable[[Any], Any], str, Callable[[Any, Any], bool], Any]] = []

    for field_name, expected_value in tests.items():
        if field_name == 'message' and remove_colors:
            def field_getter(record, name): return remove_color(getattr(record, name))

        elif field_name.startswith('details_'):
            field_name = field_name.replace('details_', '')
            def field_getter(record, name): return getattr(record.details, name, None)

        elif field_name == 'message':
            def field_getter(record, name): return remove_color(getattr(record, name))

        else:
            def field_getter(record, name): return getattr(record, name)

        # Special case: if the expected value is a pattern matching instance, it represents a
        # regular expression. We don't modify the field name and "expected" value, but the
        # function will be a custom lambda calling proper `re` method.
        if isinstance(expected_value, PatternMatching):
            operators.append((
                field_getter,
                field_name,
                lambda a, b: a.method(b) is not None,
                expected_value
                ))

            continue

        # Python's `operator` package offers operators - `==` or `!=` - in a form of functions,
        # which is exactly what we need here, so we don't have to build our own
        # `lambda a, b: a == b`. We might use more than just `eq` in the future, so let's start
        # with `operator` right away.

        operators.append((
            field_getter,
            field_name,
            operator.eq,
            expected_value
            ))

    # Given a logging record, apply all field/operator/value triplets, and make sure all match the
    # actual record properties.
    def _cmp(record: logging.LogRecord) -> bool:
        return all(op(expected_value, field_getter(record, field_name))
                   for field_getter, field_name, op, expected_value in operators)

    # Final step: apply our "make sure field/operator/value triplets match given record" to each
    # and every record, and reduce per-record results into a single answer. By default, `any` is
    # used which means that any record matching all field/operator/value triples yield the final
    # "yes, such a record exists".
    outcome = evaluator([_cmp(record) for record in caplog.records])

    def _report(message: str) -> None:
        formatted_fields = [
            f'    {field} == {value}'
            for field, value in tests.items()
            ]

        for record in caplog.records:
            for field_getter, field_name, op, expected_value in operators:
                print(f'field={field_name}',
                      f'current=>>>{field_getter(record, field_name)}<<<',
                      f'expected=>>>{expected_value}<<<',
                      f'comparison={op(expected_value, field_getter(record, field_name))}')

        pytest.fail(f"""
{message}:

{chr(10).join(formatted_fields)}
""")

    # Cannot find log record with these properties

    if not_present is False and not outcome:
        _report('Could not find log record with these properties when it should exist')

    elif not_present is True and outcome:
        _report('Found log record with these properties when it should not exist')


def assert_log(
        caplog: _pytest.logging.LogCaptureFixture,
        evaluator: Callable[[Iterable[Any]], bool] = any,
        remove_colors: bool = False,
        **tests: Any
        ) -> None:
    _assert_log(
        caplog,
        evaluator=evaluator,
        remove_colors=remove_colors,
        not_present=False,
        **tests)


def assert_not_log(
        caplog: _pytest.logging.LogCaptureFixture,
        evaluator: Callable[[Iterable[Any]], bool] = any,
        remove_colors: bool = False,
        **tests: Any
        ) -> None:
    _assert_log(
        caplog,
        evaluator=evaluator,
        remove_colors=remove_colors,
        not_present=True,
        **tests)



from typing import Any, Optional

import tmt.base
import tmt.export
import tmt.utils
from tmt.utils import Path


@tmt.base.FmfId.provides_export('template')
@tmt.base.Test.provides_export('template')
@tmt.base.Plan.provides_export('template')
@tmt.base.Story.provides_export('template')
class TemplateExporter(tmt.export.ExportPlugin):
    @classmethod
    def render_template(
            cls,
            *,
            template_filepath: Optional[Path] = None,
            default_template_filename: str,
            keys: Optional[list[str]] = None,
            **variables: Any
            ) -> str:
        return tmt.utils.templates.render_template_file(
            template_filepath or tmt.export.TEMPLATES_DIRECTORY / default_template_filename,
            KEYS=keys,
            **variables
            )

    @classmethod
    def export_fmfid_collection(cls,
                                fmf_ids: list[tmt.base.FmfId],
                                keys: Optional[list[str]] = None,
                                template: Optional[Path] = None,
                                **kwargs: Any) -> str:
        return '\n\n'.join([
            cls.render_template(
                template_filepath=template,
                default_template_filename='default-fmfid.j2',
                keys=keys,
                FMF_ID=fmf_id)
            for fmf_id in fmf_ids
            ])

    @classmethod
    def export_test_collection(cls,
                               tests: list[tmt.base.Test],
                               keys: Optional[list[str]] = None,
                               template: Optional[Path] = None,
                               **kwargs: Any) -> str:
        return '\n\n'.join([
            cls.render_template(
                template_filepath=template,
                default_template_filename='default-test.j2',
                keys=keys,
                TEST=test)
            for test in tests
            ])

    @classmethod
    def export_plan_collection(cls,
                               plans: list[tmt.base.Plan],
                               keys: Optional[list[str]] = None,
                               template: Optional[Path] = None,
                               **kwargs: Any) -> str:
        return '\n\n'.join([
            cls.render_template(
                template_filepath=template,
                default_template_filename='default-plan.j2',
                keys=keys,
                PLAN=plan)
            for plan in plans
            ])

    @classmethod
    def export_story_collection(cls,
                                stories: list[tmt.base.Story],
                                keys: Optional[list[str]] = None,
                                template: Optional[Path] = None,
                                include_title: bool = True,
                                **kwargs: Any) -> str:
        return '\n\n'.join([
            cls.render_template(
                template_filepath=template,
                default_template_filename='default-story.j2',
                keys=keys,
                STORY=story,
                INCLUDE_TITLE=include_title)
            for story in stories
            ])



import dataclasses
import functools
import re
import sys
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Callable, Optional, TypedDict, cast, overload

from jinja2 import FileSystemLoader, select_autoescape

import tmt
import tmt.base
import tmt.log
import tmt.options
import tmt.result
import tmt.steps
import tmt.steps.report
import tmt.utils
from tmt.plugins import ModuleImporter
from tmt.result import ResultOutcome
from tmt.utils import Path, field
from tmt.utils.templates import default_template_environment, render_template_file

if TYPE_CHECKING:
    import lxml

    from tmt._compat.typing import TypeAlias

    XMLElement: TypeAlias = Any


DEFAULT_NAME = 'junit.xml'
DEFAULT_FLAVOR_NAME = 'default'
CUSTOM_FLAVOR_NAME = 'custom'

# Relative path to tmt junit template directory.
DEFAULT_TEMPLATE_DIR = Path('steps/report/junit/templates/')

# ignore[unused-ignore]: Pyright would report that "module cannot be
# used as a type", and it would be correct. On the other hand, it works,
# and both mypy and pyright are able to propagate the essence of a given
# module through `ModuleImporter` that, eventually, the module object
# returned by the importer does have all expected members.
#
# The error message does not have its own code, but simple `type: ignore`
# is enough to suppress it. And then mypy complains about an unused
# ignore, hence `unused-ignore` code, leading to apparently confusing
# directive.
import_lxml: ModuleImporter['lxml'] = ModuleImporter(  # type: ignore[valid-type]
    'lxml',
    tmt.utils.ReportError,
    "Missing 'lxml', fixable by 'pip install tmt[report-junit]'.")


@overload
def _duration_to_seconds_filter(duration: str) -> int: pass


@overload
def _duration_to_seconds_filter(duration: None) -> None: pass


def _duration_to_seconds_filter(duration: Optional[str]) -> Optional[int]:
    """ Convert valid duration string in to seconds """
    if duration is None:
        return None
    try:
        h, m, s = duration.split(':')
        return int(h) * 3600 + int(m) * 60 + int(s)
    except Exception as error:
        raise tmt.utils.ReportError(f"Malformed duration '{duration}'.") from error


def _escape_control_chars_filter(func: Callable[[str], str]) -> Callable[[str], str]:
    """ Wrap the escape filter function and escape ASCII chosen control characters """

    def wrapper(value: str) -> str:
        # Define unicode characters which are not allowed in the XML and need to be removed.
        # - https://www.w3.org/TR/REC-xml/#NT-Char
        # - https://github.com/kyrus/python-junit-xml/blob/4bd08a272f059998cedf9b7779f944d49eba13a6/junit_xml/__init__.py#L325
        illegal_chars = [
            (0x00, 0x08),
            (0x0B, 0x1F),
            (0x7F, 0x84),
            (0x86, 0x9F),
            (0xD800, 0xDFFF),
            (0xFDD0, 0xFDDF),
            (0xFFFE, 0xFFFF),
            (0x1FFFE, 0x1FFFF),
            (0x2FFFE, 0x2FFFF),
            (0x3FFFE, 0x3FFFF),
            (0x4FFFE, 0x4FFFF),
            (0x5FFFE, 0x5FFFF),
            (0x6FFFE, 0x6FFFF),
            (0x7FFFE, 0x7FFFF),
            (0x8FFFE, 0x8FFFF),
            (0x9FFFE, 0x9FFFF),
            (0xAFFFE, 0xAFFFF),
            (0xBFFFE, 0xBFFFF),
            (0xCFFFE, 0xCFFFF),
            (0xDFFFE, 0xDFFFF),
            (0xEFFFE, 0xEFFFF),
            (0xFFFFE, 0xFFFFF),
            (0x10FFFE, 0x10FFFF)]

        illegal_ranges = [
            f"{chr(low)}-{chr(high)}" for low, high in illegal_chars if low < sys.maxunicode]

        illegal_regex = re.compile("[{}]".format("".join(illegal_ranges)))
        escaped_value = illegal_regex.sub("", value)

        # It's important to call the parent `func` at the end of the wrapper, otherwise the
        # jinja autoescape doesn't work correctly (some chars like '&' are escaped two times).
        return func(escaped_value)

    return wrapper


class ImplementProperties:
    """
    Define a properties attribute.

    This class can be used to easily add properties attribute by inheriting it.
    """

    class PropertyDict(TypedDict):
        """ Defines a property dict, which gets propagated into the final template properties. """

        name: str
        value: str

    def __init__(self) -> None:
        self._properties: dict[str, str] = {}

    @property
    def properties(self) -> list[PropertyDict]:
        return [{'name': k, 'value': v} for k, v in self._properties.items()]

    @properties.setter
    def properties(self, keyval: dict[str, Optional[str]]) -> None:
        self._properties = {k: v for k, v in keyval.items() if v is not None}


class ResultWrapper(ImplementProperties):
    """
    The context wrapper for :py:class:`tmt.Result`.

    Adds possibility to wrap the :py:class:`tmt.Result` and dynamically add more attributes which
    get available inside the template context.
    """

    def __init__(self, wrapped: tmt.Result) -> None:
        super().__init__()
        self._wrapped = wrapped

    def __getattr__(self, name: str) -> Any:
        """ Returns an attribute of a wrapped ``tmt.Result`` instance """
        return getattr(self._wrapped, name)


class ResultsContext(ImplementProperties):
    """
    The results context for Jinja templates.

    A class which keeps the results context (especially the result summary) for JUnit template. It
    wraps all the :py:class:`tmt.Result` instances into the :py:class:`ResultWrapper`.
    """

    def __init__(self, results: list[tmt.Result]) -> None:
        super().__init__()

        # Decorate all the tmt.Results with more attributes
        self._results: list[ResultWrapper] = [ResultWrapper(r) for r in results]

    def __iter__(self) -> Iterator[ResultWrapper]:
        """ Possibility to iterate over results by iterating an instance """
        return iter(self._results)

    def __len__(self) -> int:
        """ Returns the number of results """
        return len(self._results)

    @functools.cached_property
    def passed(self) -> list[ResultWrapper]:
        """ Returns results of passed tests """
        return [r for r in self._results if r.result == ResultOutcome.PASS]

    @functools.cached_property
    def skipped(self) -> list[ResultWrapper]:
        """ Returns results of skipped tests """
        return [r for r in self._results if r.result == ResultOutcome.INFO]

    @functools.cached_property
    def failed(self) -> list[ResultWrapper]:
        """ Returns results of failed tests """
        return [r for r in self._results if r.result == ResultOutcome.FAIL]

    @functools.cached_property
    def errored(self) -> list[ResultWrapper]:
        """ Returns results of tests with error/warn outcome """
        return [r for r in self._results if r.result in [
                ResultOutcome.ERROR,
                ResultOutcome.WARN]]

    @functools.cached_property
    def duration(self) -> int:
        """ Returns the total duration of all tests in seconds """
        # cast: mypy does not understand the proxy-ness of `ResultWrapper`. `r.duration`
        # will exists, therefore adding a `cast` to convince mypy the list is pretty much
        # nothing but the list of results.
        return sum(_duration_to_seconds_filter(r.duration)
                   or 0 for r in cast(list[tmt.Result], self._results))


def make_junit_xml(
        phase: tmt.steps.report.ReportPlugin[Any],
        flavor: str = DEFAULT_FLAVOR_NAME,
        template_path: Optional[Path] = None,
        include_output_log: bool = True,
        prettify: bool = True,
        results_context: Optional[ResultsContext] = None,
        **extra_variables: Any
        ) -> str:
    """
    Create JUnit XML file and return the data.

    :param phase: instance of a :py:class:`tmt.steps.report.ReportPlugin`.
    :param flavor: name of a JUnit flavor to generate.
    :param template_path: if set, the provided template will be used instead of a pre-defined
        flavor template. In this case, the ``flavor`` must be set to ``custom`` value.
    :param include_output_log: if enabled, the ``<system-out>`` tags are included in the final
        template output.
    :param prettify: allows to control the XML pretty print.
    :param results_context: if set, the provided :py:class:`ResultsContext` is used in a template.
    :param extra_variables: if set, these variables get propagated into the Jinja template.
    """

    # Get the template context for tmt results
    results_context = results_context or ResultsContext(phase.step.plan.execute.results())

    # Prepare the template environment
    environment = default_template_environment()

    template_path = template_path or tmt.utils.resource_files(
        DEFAULT_TEMPLATE_DIR / Path(f'{flavor}.xml.j2'))

    # Use a FileSystemLoader for a non-custom flavor
    if flavor != CUSTOM_FLAVOR_NAME:
        environment.loader = FileSystemLoader(
            searchpath=tmt.utils.resource_files(DEFAULT_TEMPLATE_DIR))

    def _read_log_filter(log: Path) -> str:
        """ Read the contents of a given result log """
        if not log:
            return ''

        try:
            return str(phase.step.plan.execute.read(log))
        except tmt.utils.FileError:
            return ''

    environment.filters.update({
        'read_log': _read_log_filter,
        'duration_to_seconds': _duration_to_seconds_filter,
        'failures': tmt.result.Result.failures,

        # Use a wrapper function to also _escape_control_chars.
        'e': _escape_control_chars_filter(environment.filters['e']),
        })

    # Explicitly enable the autoescape because it's disabled by default by tmt.
    # See /teemtee/tmt/issues/2873 for more info.
    environment.autoescape = select_autoescape(enabled_extensions=('xml.j2'))

    xml_data = render_template_file(
        template_path,
        environment,
        RESULTS=results_context,
        PLAN=phase.step.plan,
        INCLUDE_OUTPUT_LOG=include_output_log,
        **extra_variables)

    # Try to use lxml to check the flavor XML schema and prettify the final XML
    # output.
    try:
        from lxml import etree
    except ImportError:
        phase.warn(
            "Install 'tmt[report-junit]' to support neater JUnit XML output and the XML schema "
            "validation against the XSD.")
        return xml_data

    xml_parser_kwargs: dict[str, Any] = {
        'remove_blank_text': prettify,
        'schema': None,
        }

    # The schema check must be done only for a non-custom JUnit flavors
    if flavor != CUSTOM_FLAVOR_NAME:
        xsd_schema_path = Path(tmt.utils.resource_files(
            Path(f'steps/report/junit/schemas/{flavor}.xsd')))

        schema_root: XMLElement = etree.XML(xsd_schema_path.read_bytes())
        xml_parser_kwargs['schema'] = etree.XMLSchema(schema_root)
    else:
        phase.warn(
            f"The '{CUSTOM_FLAVOR_NAME}' JUnit flavor is used, you are solely responsible "
            "for the validity of the XML schema.")

        if prettify:
            phase.warn(f"The pretty print is always disabled for '{CUSTOM_FLAVOR_NAME}' JUnit "
                       "flavor.")

        xml_parser_kwargs['remove_blank_text'] = prettify = False

    xml_parser = etree.XMLParser(**xml_parser_kwargs)
    try:
        # S320: Parsing of untrusted data is known to be vulnerable to XML
        # attacks.
        tree_root: XMLElement = etree.fromstring(xml_data, xml_parser)  # noqa: S320

    except etree.XMLSyntaxError as e:
        phase.warn(
            'The generated XML output is not a valid XML file or it is not valid against the '
            'XSD schema.')

        if flavor != CUSTOM_FLAVOR_NAME:
            phase.warn('Please, report this problem to project maintainers.')

        for err in e.error_log:
            phase.warn(str(err))

        # Return the prettified XML without checking the XSD
        del xml_parser_kwargs['schema']

        xml_parser = etree.XMLParser(**xml_parser_kwargs)

        try:
            tree_root = etree.fromstring(xml_data, xml_parser)  # noqa: S320
        except etree.XMLSyntaxError as error:
            phase.verbose('rendered XML', xml_data, 'red')
            raise tmt.utils.ReportError(
                'The generated XML output is not a valid XML file. Use `--verbose` argument '
                'to show the output.') from error

    # Do not be fooled by the `encoding` parameter: even with `utf-8`,
    # `tostring()` will still return bytes. `unicode`, on the other
    # hand, would give us a string, but then, no XML declaration.
    # So, we get bytes, and we need to apply `decode()` on our own.
    xml_output: bytes = etree.tostring(
        tree_root,
        xml_declaration=True,
        pretty_print=prettify,
        # The 'utf-8' encoding must be used instead of 'unicode', otherwise
        # the XML declaration is not included in the output.
        encoding='utf-8')

    return str(xml_output.decode('utf-8'))


@dataclasses.dataclass
class ReportJUnitData(tmt.steps.report.ReportStepData):
    file: Optional[Path] = field(
        default=None,
        option='--file',
        metavar='PATH',
        help='Path to the file to store JUnit to.',
        normalize=tmt.utils.normalize_path)

    flavor: str = field(
        default=DEFAULT_FLAVOR_NAME,
        option='--flavor',
        choices=[DEFAULT_FLAVOR_NAME, CUSTOM_FLAVOR_NAME],
        help='Name of a JUnit flavor to generate.')

    template_path: Optional[Path] = field(
        default=None,
        option='--template-path',
        metavar='TEMPLATE_PATH',
        help='Path to a custom template file to use for JUnit creation.',
        normalize=tmt.utils.normalize_path)

    prettify: bool = field(
        default=True,
        option=('--prettify / --no-prettify'),
        is_flag=True,
        show_default=True,
        help=f"""
            Enable the XML pretty print for generated JUnit file. This option is always disabled
            for '{CUSTOM_FLAVOR_NAME}' template flavor.
            """
        )

    include_output_log: bool = field(
        default=True,
        option=('--include-output-log / --no-include-output-log'),
        is_flag=True,
        show_default=True,
        help='Include full standard output in resulting xml file.')


@tmt.steps.provides_method('junit')
class ReportJUnit(tmt.steps.report.ReportPlugin[ReportJUnitData]):
    """
    Save test results in chosen JUnit flavor format.

    When flavor is set to custom, the ``template-path`` with a path to a custom template must be
    provided.

    When ``file`` is not specified, output is written into a file named ``junit.xml`` located in
    the current workdir.
    """

    _data_class = ReportJUnitData

    def check_options(self) -> None:
        """ Check the module options """

        if self.data.flavor == 'custom' and not self.data.template_path:
            raise tmt.utils.ReportError(
                "The 'custom' flavor requires the '--template-path' argument.")

        if self.data.flavor != 'custom' and self.data.template_path:
            raise tmt.utils.ReportError(
                "The '--template-path' can be used only with '--flavor=custom'.")

    def prune(self, logger: tmt.log.Logger) -> None:
        """ Do not prune generated junit report """

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """ Read executed tests and write junit """
        super().go(logger=logger)

        self.check_options()

        assert self.workdir is not None
        f_path = self.data.file or self.workdir / DEFAULT_NAME

        xml_data = make_junit_xml(
            phase=self,
            flavor=self.data.flavor,
            template_path=self.data.template_path,
            include_output_log=self.data.include_output_log,
            prettify=self.data.prettify)
        try:
            f_path.write_text(xml_data)
            self.info('output', f_path, 'yellow')
        except Exception as error:
            raise tmt.utils.ReportError(f"Failed to write the output '{f_path}'.") from error



import textwrap

import pytest

import tmt.utils
from tests.unit.test_hardware import FULL_HARDWARE_REQUIREMENTS
from tmt.hardware import (
    Hardware,
    Operator,
    _parse_cpu,
    _parse_disk,
    _parse_hostname,
    _parse_iommu,
    _parse_location,
    _parse_memory,
    _parse_system,
    _parse_tpm,
    _parse_virtualization,
    _parse_zcrypt,
    )
from tmt.log import Logger
from tmt.steps.provision.mrack import (
    _CONSTRAINT_TRANSFORMERS,
    constraint_to_beaker_filter,
    operator_to_beaker_op,
    )


@pytest.mark.parametrize(
    ('operator', 'value', 'expected'),
    [
        (Operator.EQ, 'foo', ('==', 'foo', False)),
        (Operator.NEQ, 'foo', ('!=', 'foo', False)),
        (Operator.GT, 'foo', ('>', 'foo', False)),
        (Operator.GTE, 'foo', ('>=', 'foo', False)),
        (Operator.LT, 'foo', ('<', 'foo', False)),
        (Operator.LTE, 'foo', ('<=', 'foo', False)),
        (Operator.MATCH, 'f.+o.*o', ('like', 'f%o%o', False)),
        (Operator.NOTMATCH, 'f.+o.*o', ('like', 'f%o%o', True))
        ]
    )
def test_operator_to_beaker_op(
        operator: Operator,
        value: str,
        expected: tuple[str, str, bool]) -> None:
    assert operator_to_beaker_op(operator, value) == expected


def test_maximal_constraint(root_logger: Logger) -> None:
    hw = Hardware.from_spec(tmt.utils.yaml_to_dict(textwrap.dedent(FULL_HARDWARE_REQUIREMENTS)))
    assert hw.constraint is not None

    result = constraint_to_beaker_filter(hw.constraint, root_logger)

    assert result.to_mrack() == {
        'and': [
            {'pool': {
                '_op': '!=',
                '_value': 'foo.*',
                },
             },
            {'or': []},
            {
                'and': [
                    {'or': []},
                    {'or': []}
                    ]
                },
            {
                'and': [
                    {
                        'cpu': {
                            'processors': {
                                '_op': '>',
                                '_value': '8'
                                },
                            },
                        },
                    {
                        'or': [],
                        },
                    {
                        'cpu': {
                            'cores': {
                                '_op': '==',
                                '_value': '2'
                                },
                            },
                        },
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {
                        'cpu': {
                            'model': {
                                '_op': '==',
                                '_value': '62'
                                }
                            }
                        },
                    {'or': []},
                    {'or': []},
                    {
                        'cpu': {
                            'stepping': {
                                '_op': '!=',
                                '_value': '10',
                                },
                            },
                        },
                    {'or': []},
                    {'or': []},
                    {
                        'not':
                            {
                                'cpu': {
                                    'model_name': {
                                        '_op': 'like',
                                        '_value': 'Haswell'
                                        }
                                    }
                                },
                        },
                    {
                        'cpu': {
                            'vendor': {
                                '_op': 'like',
                                '_value': 'Intel%',
                                },
                            },
                        },
                    {
                        'and': [
                            {
                                'cpu': {
                                    'flag': {
                                        '_op': '==',
                                        '_value': 'avx'
                                        }
                                    }
                                },
                            {
                                'cpu': {
                                    'flag': {
                                        '_op': '==',
                                        '_value': 'avx2'
                                        }
                                    }
                                },
                            {
                                'cpu': {
                                    'flag': {
                                        '_op': '!=',
                                        '_value': 'smep'
                                        }
                                    }
                                },
                            ]
                        },
                    {
                        'cpu': {
                            'hyper': {
                                '_op': '==',
                                '_value': 'True'
                                }
                            }
                        },
                    ]
                },
            {
                'and': [
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {'or': []}
                    ]
                },
            {
                'and': [
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {'or': []}
                    ]
                },
            {
                'system': {
                    'memory': {
                        '_op': '==',
                        '_value': '8192'
                        }
                    }
                },
            {
                'and': [
                    {
                        'and': [
                            {
                                'disk': {
                                    'size': {
                                        '_op': '==',
                                        '_value': '42949672960'
                                        }
                                    }
                                },
                            {
                                'disk': {
                                    'phys_sector_size': {
                                        '_op': '==',
                                        '_value': '4096 B'
                                        }
                                    }
                                },
                            {
                                'disk': {
                                    'model': {
                                        '_op': 'like',
                                        '_value': 'WD 100G%'
                                        }
                                    }
                                }
                            ]
                        },
                    {
                        'and': [
                            {
                                'disk': {
                                    'size': {
                                        '_op': '==',
                                        '_value': '128849018880'
                                        }
                                    }
                                },
                            {
                                'disk': {
                                    'sector_size': {
                                        '_op': '==',
                                        '_value': '512 B'
                                        }
                                    }
                                },
                            {
                                'key_value': {
                                    '_key': 'BOOTDISK',
                                    '_op': '==',
                                    '_value': 'virtblk'
                                    }
                                }
                            ]
                        },
                    ]
                },
            {
                'and': [
                    {
                        'and': [
                            {'or': []},
                            {'or': []},
                            {'or': []},
                            {'or': []},
                            {'or': []},
                            {'or': []}
                            ]
                        },
                    {'or': []}
                    ],
                },
            {
                'hostname': {
                    '_op': 'like',
                    '_value': '%.foo.redhat.com'
                    }
                },
            {
                'labcontroller': {
                    '_op': '!=',
                    '_value': 'lab-1.bar.redhat.com'
                    }
                },
            {
                'and': [
                    {'or': []},
                    {'or': []},
                    {'or': []},
                    {
                        'system': {
                            'numanodes': {
                                '_op': '<',
                                '_value': '4',
                                },
                            },
                        },
                    {'or': []},
                    ]
                },
            {
                'key_value': {
                    '_key': 'TPM',
                    '_op': '==',
                    '_value': '2.0',
                    },
                },

            {
                'and': [
                    {
                        'system': {
                            'hypervisor': {
                                '_op': '==',
                                '_value': ''
                                }
                            }
                        },
                    {
                        'or': []
                        },
                    {
                        'system': {
                            'hypervisor': {
                                '_op': 'like',
                                '_value': 'xen'
                                }
                            }
                        }
                    ]
                },
            {
                'and': [
                    {
                        'system': {
                            'key_value': {
                                '_key': 'ZCRYPT_MODEL',
                                '_op': '==',
                                '_value': 'CEX8C'
                                }
                            }
                        },
                    {
                        'system': {
                            'key_value': {
                                '_key': 'ZCRYPT_MODE',
                                '_op': '==',
                                '_value': 'CCA'
                                }
                            }
                        }
                    ]
                },

            ]
        }


def test_cpu_model(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.model'](_parse_cpu({'model': '79'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'model': {
                '_op': '==',
                '_value': '79'
                }
            }
        }


def test_cpu_processors(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.processors'](
        _parse_cpu({'processors': '79'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'processors': {
                '_op': '==',
                '_value': '79'
                }
            }
        }


def test_cpu_cores(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.cores'](
        _parse_cpu({'cores': '2'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'cores': {
                '_op': '==',
                '_value': '2'
                }
            }
        }


def test_cpu_stepping(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.stepping'](
        _parse_cpu({'stepping': '10'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'stepping': {
                '_op': '==',
                '_value': '10'
                }
            }
        }


def test_cpu_vendor_name(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': 'GenuineIntel'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'vendor': {
                '_op': '==',
                '_value': 'GenuineIntel'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '!= GenuineIntel'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'vendor': {
                '_op': '!=',
                '_value': 'GenuineIntel'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '~ .*Intel'}), root_logger)

    assert result.to_mrack() == {
        'cpu': {
            'vendor': {
                '_op': 'like',
                '_value': '%Intel'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '!~ .*Intel'}), root_logger)

    assert result.to_mrack() == {
        'not': {
            'cpu': {
                'vendor': {
                    '_op': 'like',
                    '_value': '%Intel'
                    }
                }
            }
        }


def test_disk_driver(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': 'mpt3sas'}, 1), root_logger)

    assert result.to_mrack() == {
        'key_value': {
            '_key': 'BOOTDISK',
            '_op': '==',
            '_value': 'mpt3sas'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '!= mpt3sas'}, 1), root_logger)

    assert result.to_mrack() == {
        'key_value': {
            '_key': 'BOOTDISK',
            '_op': '!=',
            '_value': 'mpt3sas'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '~ mpt3.*'}, 1), root_logger)

    assert result.to_mrack() == {
        'key_value': {
            '_key': 'BOOTDISK',
            '_op': 'like',
            '_value': 'mpt3%'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '!~ mpt3.*'}, 1), root_logger)

    assert result.to_mrack() == {
        'not': {
            'key_value': {
                '_key': 'BOOTDISK',
                '_op': 'like',
                '_value': 'mpt3%'
                }
            }
        }


def test_disk_size(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.size'](
        _parse_disk({'size': '>= 40 GiB'}, 1), root_logger)

    assert result.to_mrack() == {
        'disk': {
            'size': {
                '_op': '>=',
                '_value': '42949672960'
                }
            }
        }


def test_disk_model_name(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': 'PERC H310'}, 1), root_logger)

    assert result.to_mrack() == {
        'disk': {
            'model': {
                '_op': '==',
                '_value': 'PERC H310'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '!= PERC H310'}, 1), root_logger)

    assert result.to_mrack() == {
        'disk': {
            'model': {
                '_op': '!=',
                '_value': 'PERC H310'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '~ PERC.*'}, 1), root_logger)

    assert result.to_mrack() == {
        'disk': {
            'model': {
                '_op': 'like',
                '_value': 'PERC%'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '!~ PERC.*'}, 1), root_logger)

    assert result.to_mrack() == {
        'not': {
            'disk': {
                'model': {
                    '_op': 'like',
                    '_value': 'PERC%'
                    }
                }
            }
        }


def test_memory(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['memory'](_parse_memory({'memory': '>= 4 GiB'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'memory': {
                '_op': '>=',
                '_value': '4096'
                }
            }
        }


def test_hostname(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': 'foo.dot.com'}), root_logger)

    assert result.to_mrack() == {
        'hostname': {
            '_op': '==',
            '_value': 'foo.dot.com'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': '~ foo.*.dot.+.com'}), root_logger)

    assert result.to_mrack() == {
        'hostname': {
            '_op': 'like',
            '_value': 'foo%.dot%.com'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': '!~ foo.*.dot.+.com'}), root_logger)

    assert result.to_mrack() == {
        'not': {
            'hostname': {
                '_op': 'like',
                '_value': 'foo%.dot%.com'
                }
            }
        }


def test_virtualization_is_virtualized(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['virtualization.is_virtualized'](
        _parse_virtualization({'is-virtualized': True}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'hypervisor': {
                '_op': '!=',
                '_value': ''
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['virtualization.is_virtualized'](
        _parse_virtualization({'is-virtualized': False}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'hypervisor': {
                '_op': '==',
                '_value': ''
                }
            }
        }


def test_virtualization_hypervisor(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "~ kvm"}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'hypervisor': {
                '_op': 'like',
                '_value': 'kvm'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "!~ kvm"}), root_logger)

    assert result.to_mrack() == {
        'not': {
            'system': {
                'hypervisor': {
                    '_op': 'like',
                    '_value': 'kvm'
                    }
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "kvm"}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'hypervisor': {
                '_op': '==',
                '_value': 'kvm'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "!= kvm"}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'hypervisor': {
                '_op': '!=',
                '_value': 'kvm'
                }
            }
        }


def test_zcrypt_adapter(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': 'CEX8C'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODEL',
                '_op': '==',
                '_value': 'CEX8C'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '!= CEX8C'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODEL',
                '_op': '!=',
                '_value': 'CEX8C'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '~ CEX.*'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODEL',
                '_op': 'like',
                '_value': 'CEX%'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '!~ CEX.*'}), root_logger)

    assert result.to_mrack() == {
        'not': {
            'system': {
                'key_value': {
                    '_key': 'ZCRYPT_MODEL',
                    '_op': 'like',
                    '_value': 'CEX%'
                    }
                }
            }
        }


def test_zcrypt_mode(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': 'CCA'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODE',
                '_op': '==',
                '_value': 'CCA'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '!= CCA'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODE',
                '_op': '!=',
                '_value': 'CCA'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '~ C.*A'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'key_value': {
                '_key': 'ZCRYPT_MODE',
                '_op': 'like',
                '_value': 'C%A'
                }
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '!~ C.*A'}), root_logger)

    assert result.to_mrack() == {
        'not': {
            'system': {
                'key_value': {
                    '_key': 'ZCRYPT_MODE',
                    '_op': 'like',
                    '_value': 'C%A'
                    }
                }
            }
        }


def test_iommu_is_supported(root_logger: Logger) -> None:

    for value in True, False:

        result = _CONSTRAINT_TRANSFORMERS['iommu.is_supported'](
            _parse_iommu({"is-supported": value}), root_logger)

        assert result.to_mrack() == {
            'key_value': {
                '_key': 'VIRT_IOMMU',
                '_op': '==',
                '_value': str(int(value))
                }
            }


def test_location_lab_controller(root_logger: Logger) -> None:

    result = _CONSTRAINT_TRANSFORMERS['location.lab_controller'](
        _parse_location({"lab-controller": "lab-01.bar.redhat.com"}), root_logger)

    assert result.to_mrack() == {
        'labcontroller': {
            '_op': '==',
            '_value': 'lab-01.bar.redhat.com'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['location.lab_controller'](
        _parse_location({"lab-controller": "!= lab-01.bar.redhat.com"}), root_logger)

    assert result.to_mrack() == {
        'labcontroller': {
            '_op': '!=',
            '_value': 'lab-01.bar.redhat.com'
            }
        }


def test_tpm_version(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['tpm.version'](
        _parse_tpm({'version': '2.0'}), root_logger)

    assert result.to_mrack() == {
        'key_value': {
            '_key': 'TPM',
            '_op': '==',
            '_value': '2.0'
            }
        }

    result = _CONSTRAINT_TRANSFORMERS['tpm.version'](
        _parse_tpm({'version': '!= 2.0'}), root_logger)

    assert result.to_mrack() == {
        'key_value': {
            '_key': 'TPM',
            '_op': '!=',
            '_value': '2.0'
            }
        }


def test_system_numa_nodes(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['system.numa_nodes'](
        _parse_system({'numa-nodes': '2'}), root_logger)

    assert result.to_mrack() == {
        'system': {
            'numanodes': {
                '_op': '==',
                '_value': '2'
                }
            }
        }



from typing import Any, Optional

import tmt.base
import tmt.export
import tmt.export.template
import tmt.utils
from tmt.utils import Path


@tmt.base.Story.provides_export('rst')
class RestructuredExporter(tmt.export.ExportPlugin):
    @classmethod
    def export_story(cls,
                     story: tmt.base.Story,
                     keys: Optional[list[str]] = None,
                     template: Optional[Path] = None,
                     include_title: bool = True) -> str:
        return tmt.export.template.TemplateExporter.render_template(
            template_filepath=template,
            default_template_filename='default-story.rst.j2',
            keys=keys,
            STORY=story,
            INCLUDE_TITLE=include_title)

    @classmethod
    def export_story_collection(cls,
                                stories: list[tmt.base.Story],
                                keys: Optional[list[str]] = None,
                                template: Optional[Path] = None,
                                include_title: bool = True,
                                **kwargs: Any) -> str:
        return '\n\n'.join([
            cls.export_story(story, keys=keys, template=template, include_title=include_title)
            for story in stories])
